# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import re
import sys
sys.path.insert(0, os.path.abspath('.'))
sys.path.append('skrmt')


# -- Project information -----------------------------------------------------

project = 'scikit-rmt'
copyright = '2021, Alejandro Santorum Varela'
author = 'Alejandro Santorum Varela'

# The full version, including alpha/beta/rc tags
VERSION_FILEPATH = "skrmt/_version.py"
with open(
    os.path.join(os.path.dirname(__file__), VERSION_FILEPATH),
    "r",
) as version_file:
    version_file_text = version_file.read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    match = re.search(VSRE, version_file_text, re.M)
    if match:
        release = match.group(1)
    else:
        raise RuntimeError(f"Unable to find version in {VERSION_FILEPATH}.")

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx_gallery.gen_gallery'
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Add any paths that contain templates here, relative to this directory.
#templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'venv/*']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# tutorial_list = [
#     "plot_introduction.py",
# ]

sphinx_gallery_conf = {
    'examples_dirs': ['examples', 'tutorial'],
    'gallery_dirs': ['auto_examples', 'auto_tutorial'],
}